from typing import Dict, Any


def add_opik_tracer_to_params(params: Dict[str, Any]) -> Dict[str, Any]:
    from opik.integrations.langchain import OpikTracer

    opik_tracer = OpikTracer()

    if "config" not in params:
        params["config"] = {}

    if "callbacks" not in params["config"]:
        params["config"]["callbacks"] = [opik_tracer]
        return params

    callbacks = params["config"]["callbacks"]
    for callback in callbacks:
        if isinstance(callback, opik_tracer.__class__):
            return params

    callbacks.append(opik_tracer)

    return params



import inspect
from typing import Any, Dict, Optional, Protocol, Union, List

from opik.evaluation.metrics import score_result
from opik.message_processing.emulation import models


class ScorerFunctionProtocol(Protocol):
    """
    Represents a protocol defining the structure for a scorer function.

    This protocol serves as a contract for implementing scorer functions used in
    evaluating tasks. A scorer function adhering to this protocol should take
    dataset item data, task outputs, and optionally a task span model as input
    parameters and return a scoring result.
    """

    def __call__(
        self,
        dataset_item: Dict[str, Any],
        task_outputs: Dict[str, Any],
        task_span: Optional[models.SpanModel] = None,
    ) -> Union[score_result.ScoreResult, List[score_result.ScoreResult]]: ...


ScorerFunction = ScorerFunctionProtocol


EXPECTED_SCORER_FUNCTION_PARAMETERS = ["dataset_item", "task_outputs"]


def validate_scorer_function(scorer_function: ScorerFunction) -> None:
    if not callable(scorer_function):
        raise ValueError("scorer_function must be a callable function")

    parameters = inspect.signature(scorer_function).parameters
    names = set(parameters.keys())

    # Check if it has both dataset_item and task_outputs
    has_dataset_item_and_task_outputs = all(
        param in names for param in EXPECTED_SCORER_FUNCTION_PARAMETERS
    )

    # Check if it has at least one task_span parameter
    has_task_span = "task_span" in names

    if not (has_dataset_item_and_task_outputs or has_task_span):
        raise ValueError(
            f"scorer_function must have either both 'dataset_item' and 'task_outputs' parameters "
            f"or at least one 'task_span' parameter. Found parameters: {list(names)}"
        )


def has_task_span_in_parameters(scorer_function: ScorerFunction) -> bool:
    return "task_span" in inspect.signature(scorer_function).parameters



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .dataset_item import DatasetItem


class DatasetExpansionResponse(UniversalBaseModel):
    generated_samples: typing.Optional[typing.List[DatasetItem]] = pydantic.Field(default=None)
    """
    List of generated synthetic dataset items
    """

    model: typing.Optional[str] = pydantic.Field(default=None)
    """
    Model used for generation
    """

    total_generated: typing.Optional[int] = pydantic.Field(default=None)
    """
    Total number of samples generated
    """

    generation_time: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    Generation timestamp
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from typing import Optional, List, Callable

from .. import asyncio_support
from ...api_objects import opik_client
from ...api_objects.threads import threads_client
from ..metrics.conversation import conversation_thread_metric
from . import evaluation_engine, evaluation_result
from opik.rest_api import JsonListStringPublic


def evaluate_threads(
    project_name: str,
    filter_string: Optional[str],
    eval_project_name: Optional[str],
    metrics: List[conversation_thread_metric.ConversationThreadMetric],
    trace_input_transform: Callable[[JsonListStringPublic], str],
    trace_output_transform: Callable[[JsonListStringPublic], str],
    verbose: int = 1,
    num_workers: int = 8,
    max_traces_per_thread: int = 1000,
) -> evaluation_result.ThreadsEvaluationResult:
    """Evaluate conversation threads using specified metrics.

    This function evaluates conversation threads from a project using the provided metrics.
    It creates a ThreadsEvaluationEngine to fetch threads matching the filter string,
    converts them to conversation threads, applies the metrics, and logs feedback scores.

    Args:
        project_name: The name of the project containing the threads to evaluate.
        filter_string: Optional filter string to select specific threads for evaluation using Opik Query Language (OQL).
            The format is: "<COLUMN> <OPERATOR> <VALUE> [AND <COLUMN> <OPERATOR> <VALUE>]*"

            Supported columns include:
            - `id`, `name`, `created_by`, `thread_id`, `type`, `model`, `provider`: String fields with full operator support
            - `status`: String field (=, contains, not_contains only)
            - `start_time`, `end_time`: DateTime fields (use ISO 8601 format, e.g., "2024-01-01T00:00:00Z")
            - `input`, `output`: String fields for content (=, contains, not_contains only)
            - `metadata`: Dictionary field (use dot notation, e.g., "metadata.model")
            - `feedback_scores`: Numeric field (use dot notation, e.g., "feedback_scores.accuracy")
            - `tags`: List field (use "contains" operator only)
            - `usage.total_tokens`, `usage.prompt_tokens`, `usage.completion_tokens`: Numeric usage fields
            - `duration`, `number_of_messages`, `total_estimated_cost`: Numeric fields

            Examples: 'status = "inactive"', 'id = "thread_123"', 'duration > 300'
            If None, all threads in the project will be evaluated.
        eval_project_name: Optional name for the evaluation project where evaluation traces will be stored.
            If None, the same project_name will be used.
        metrics: List of ConversationThreadMetric instances to apply to each thread.
            Must contain at least one metric.
        trace_input_transform: Function to transform trace input JSON to string representation.
            This function extracts the relevant user message from your trace's input structure.
            The function receives the raw trace input as a dictionary and should return a string.

            Example: If your trace input is {"content": {"user_question": "Hello"}},
            use: lambda x: x["content"]["user_question"]

            This transformation is essential because trace inputs vary by framework, but metrics
            expect a standardized string format representing the user's message.
        trace_output_transform: Function to transform trace output JSON to string representation.
            This function extracts the relevant agent response from your trace's output structure.
            The function receives the raw trace output as a dictionary and should return a string.

            Example: If your trace output is {"response": {"text": "Hi there"}},
            use: lambda x: x["response"]["text"]

            This transformation is essential because trace outputs vary by framework, but metrics
            expect a standardized string format representing the agent's response.
        verbose: Verbosity level for progress reporting (0=silent, 1=progress).
            Default is 1.
        num_workers: Number of concurrent workers for thread evaluation.
            Default is 8.
        max_traces_per_thread: Maximum number of traces to fetch per thread.
            Default is 1000.

    Returns:
        ThreadsEvaluationResult containing evaluation scores for each thread.

    Raises:
        ValueError: If no metrics are provided.
        MetricComputationError: If no threads are found or if evaluation fails.

    Example:
        >>> from opik.evaluation import evaluate_threads
        >>> from opik.evaluation.metrics import ConversationalCoherenceMetric, UserFrustrationMetric
        >>>
        >>> # Initialize the evaluation metrics
        >>> conversation_coherence_metric = ConversationalCoherenceMetric()
        >>> user_frustration_metric = UserFrustrationMetric()
        >>>
        >>> # Run the threads evaluation
        >>> results = evaluate_threads(
        >>>     project_name="ai_team",
        >>>     filter_string='thread_id = "0197ad2a-cf5c-75af-be8b-20e8a23304fe"',
        >>>     eval_project_name="ai_team_evaluation",
        >>>     metrics=[
        >>>         conversation_coherence_metric,
        >>>         user_frustration_metric,
        >>>     ],
        >>>     trace_input_transform=lambda x: x["input"],
        >>>     trace_output_transform=lambda x: x["output"],
        >>> )
    """
    client = opik_client.get_client_cached()
    threads_client_ = threads_client.ThreadsClient(client)

    with asyncio_support.async_http_connections_expire_immediately():
        engine = evaluation_engine.ThreadsEvaluationEngine(
            client=threads_client_,
            project_name=project_name,
            number_of_workers=num_workers,
            verbose=verbose,
        )
        return engine.evaluate_threads(
            filter_string=filter_string,
            eval_project_name=eval_project_name,
            metrics=metrics,
            trace_input_transform=trace_input_transform,
            trace_output_transform=trace_output_transform,
            max_traces_per_thread=max_traces_per_thread,
        )



import inspect
import logging
from typing import Any, Dict, Optional, Callable

from .. import arguments_helpers
from ...api_objects import trace, data_helpers
from . import api_classes


LOGGER = logging.getLogger(__name__)


def extract_opik_args(
    kwargs: Dict[str, Any], func: Callable
) -> Optional[api_classes.OpikArgs]:
    """
    Extracts opik_args from kwargs and returns the parsed OpikArgs or None.

    Mutates kwargs in place by removing "opik_args" if the function does not explicitly
    declare opik_args as a parameter.

    Args:
        kwargs: Function keyword arguments
        func: The function being decorated

    Returns:
        opik_args if found, otherwise None.
    """

    # Check if a function has opik_args in its signature
    try:
        sig = inspect.signature(func)
        has_opik_args_param = "opik_args" in sig.parameters
    except (ValueError, TypeError):
        # If we can't inspect the signature, assume no opik_args parameter
        has_opik_args_param = False

    # If function explicitly has opik_args parameter, just get it without popping
    if has_opik_args_param:
        opik_args_dict = kwargs.get("opik_args", None)
    else:
        # If function doesn't have opik_args parameter, pop it from kwargs
        opik_args_dict = kwargs.pop("opik_args", None)

    if opik_args_dict is None:
        return None

    return api_classes.OpikArgs.from_dict(opik_args_dict)


def apply_opik_args_to_trace(
    opik_args: Optional[api_classes.OpikArgs],
    trace_data: Optional[trace.TraceData],
) -> None:
    """
    Apply opik_args to the trace data, including thread_id, tags, and metadata.

    Args:
        opik_args: The configuration extracted from function arguments
        trace_data: The current trace data to modify
    """
    if opik_args is None or opik_args.trace_args is None or trace_data is None:
        return

    # Handle thread_id
    if opik_args.trace_args.thread_id is not None:
        # Check for thread_id conflict
        if (
            trace_data.thread_id is not None
            and trace_data.thread_id != opik_args.trace_args.thread_id
        ):
            LOGGER.warning(
                "Trace already has thread_id='%s', but opik_args specifies thread_id='%s'. Keeping existing thread_id.",
                trace_data.thread_id,
                opik_args.trace_args.thread_id,
            )
        else:
            # Apply thread_id to trace
            trace_data.thread_id = opik_args.trace_args.thread_id

    # Apply trace tags if specified
    if opik_args.trace_args.tags:
        existing_tags = trace_data.tags or []
        merged_tags = data_helpers.merge_tags(existing_tags, opik_args.trace_args.tags)
        trace_data.tags = merged_tags

    # Apply trace metadata if specified
    if opik_args.trace_args.metadata:
        existing_metadata = trace_data.metadata or {}
        merged_metadata = data_helpers.merge_metadata(
            existing_metadata, opik_args.trace_args.metadata
        )
        trace_data.metadata = merged_metadata


def apply_opik_args_to_start_span_params(
    params: arguments_helpers.StartSpanParameters,
    opik_args: Optional[api_classes.OpikArgs],
) -> arguments_helpers.StartSpanParameters:
    """Apply opik_args to StartSpanParameters, merging tags and metadata."""
    if opik_args is None or opik_args.span_args is None:
        return params

    span_config = opik_args.span_args

    # Merge tags and metadata
    merged_tags = data_helpers.merge_tags(params.tags, span_config.tags)
    merged_metadata = data_helpers.merge_metadata(params.metadata, span_config.metadata)

    # Create updated parameters
    return arguments_helpers.StartSpanParameters(
        type=params.type,
        name=params.name,
        tags=merged_tags,
        metadata=merged_metadata,
        input=params.input,
        project_name=params.project_name,
        model=params.model,
        provider=params.provider,
    )



import abc
from typing import Any

from . import result


class Validator(abc.ABC):
    @abc.abstractmethod
    def validate(self) -> result.ValidationResult:
        pass


class RaisableValidator(Validator):
    """
    Abstract validator class that extends Validator and adds raise_if_validation_failed method.

    This is used for validators that need to raise ValidationError exceptions
    when validation fails, typically used in class initialization.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the validator.

        Subclasses can override this method with their own initialization signature.
        """
        pass

    @abc.abstractmethod
    def raise_if_validation_failed(self) -> None:
        """
        Raise a ValidationError if validation failed.

        This method should check the validation result and raise an appropriate
        ValidationError exception if validation failed.
        """
        pass



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..types.chat_completion_response import ChatCompletionResponse
from ..types.function import Function
from ..types.function_call import FunctionCall
from ..types.message import Message
from ..types.response_format import ResponseFormat
from ..types.stream_options import StreamOptions
from ..types.tool import Tool

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawChatCompletionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def create_chat_completions(
        self,
        *,
        model: typing.Optional[str] = OMIT,
        messages: typing.Optional[typing.Sequence[Message]] = OMIT,
        temperature: typing.Optional[float] = OMIT,
        top_p: typing.Optional[float] = OMIT,
        n: typing.Optional[int] = OMIT,
        stream: typing.Optional[bool] = OMIT,
        stream_options: typing.Optional[StreamOptions] = OMIT,
        stop: typing.Optional[typing.Sequence[str]] = OMIT,
        max_tokens: typing.Optional[int] = OMIT,
        max_completion_tokens: typing.Optional[int] = OMIT,
        presence_penalty: typing.Optional[float] = OMIT,
        frequency_penalty: typing.Optional[float] = OMIT,
        logit_bias: typing.Optional[typing.Dict[str, int]] = OMIT,
        user: typing.Optional[str] = OMIT,
        response_format: typing.Optional[ResponseFormat] = OMIT,
        seed: typing.Optional[int] = OMIT,
        tools: typing.Optional[typing.Sequence[Tool]] = OMIT,
        tool_choice: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        parallel_tool_calls: typing.Optional[bool] = OMIT,
        store: typing.Optional[bool] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        reasoning_effort: typing.Optional[str] = OMIT,
        service_tier: typing.Optional[str] = OMIT,
        functions: typing.Optional[typing.Sequence[Function]] = OMIT,
        function_call: typing.Optional[FunctionCall] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[ChatCompletionResponse]:
        """
        Create chat completions

        Parameters
        ----------
        model : typing.Optional[str]

        messages : typing.Optional[typing.Sequence[Message]]

        temperature : typing.Optional[float]

        top_p : typing.Optional[float]

        n : typing.Optional[int]

        stream : typing.Optional[bool]

        stream_options : typing.Optional[StreamOptions]

        stop : typing.Optional[typing.Sequence[str]]

        max_tokens : typing.Optional[int]

        max_completion_tokens : typing.Optional[int]

        presence_penalty : typing.Optional[float]

        frequency_penalty : typing.Optional[float]

        logit_bias : typing.Optional[typing.Dict[str, int]]

        user : typing.Optional[str]

        response_format : typing.Optional[ResponseFormat]

        seed : typing.Optional[int]

        tools : typing.Optional[typing.Sequence[Tool]]

        tool_choice : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]

        parallel_tool_calls : typing.Optional[bool]

        store : typing.Optional[bool]

        metadata : typing.Optional[typing.Dict[str, str]]

        reasoning_effort : typing.Optional[str]

        service_tier : typing.Optional[str]

        functions : typing.Optional[typing.Sequence[Function]]

        function_call : typing.Optional[FunctionCall]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ChatCompletionResponse]
            Chat completions response
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/chat/completions",
            method="POST",
            json={
                "model": model,
                "messages": messages,
                "temperature": temperature,
                "top_p": top_p,
                "n": n,
                "stream": stream,
                "stream_options": convert_and_respect_annotation_metadata(
                    object_=stream_options, annotation=StreamOptions, direction="write"
                ),
                "stop": stop,
                "max_tokens": max_tokens,
                "max_completion_tokens": max_completion_tokens,
                "presence_penalty": presence_penalty,
                "frequency_penalty": frequency_penalty,
                "logit_bias": logit_bias,
                "user": user,
                "response_format": convert_and_respect_annotation_metadata(
                    object_=response_format, annotation=ResponseFormat, direction="write"
                ),
                "seed": seed,
                "tools": convert_and_respect_annotation_metadata(
                    object_=tools, annotation=typing.Sequence[Tool], direction="write"
                ),
                "tool_choice": tool_choice,
                "parallel_tool_calls": parallel_tool_calls,
                "store": store,
                "metadata": metadata,
                "reasoning_effort": reasoning_effort,
                "service_tier": service_tier,
                "functions": convert_and_respect_annotation_metadata(
                    object_=functions, annotation=typing.Sequence[Function], direction="write"
                ),
                "function_call": convert_and_respect_annotation_metadata(
                    object_=function_call, annotation=FunctionCall, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ChatCompletionResponse,
                    parse_obj_as(
                        type_=ChatCompletionResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawChatCompletionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def create_chat_completions(
        self,
        *,
        model: typing.Optional[str] = OMIT,
        messages: typing.Optional[typing.Sequence[Message]] = OMIT,
        temperature: typing.Optional[float] = OMIT,
        top_p: typing.Optional[float] = OMIT,
        n: typing.Optional[int] = OMIT,
        stream: typing.Optional[bool] = OMIT,
        stream_options: typing.Optional[StreamOptions] = OMIT,
        stop: typing.Optional[typing.Sequence[str]] = OMIT,
        max_tokens: typing.Optional[int] = OMIT,
        max_completion_tokens: typing.Optional[int] = OMIT,
        presence_penalty: typing.Optional[float] = OMIT,
        frequency_penalty: typing.Optional[float] = OMIT,
        logit_bias: typing.Optional[typing.Dict[str, int]] = OMIT,
        user: typing.Optional[str] = OMIT,
        response_format: typing.Optional[ResponseFormat] = OMIT,
        seed: typing.Optional[int] = OMIT,
        tools: typing.Optional[typing.Sequence[Tool]] = OMIT,
        tool_choice: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        parallel_tool_calls: typing.Optional[bool] = OMIT,
        store: typing.Optional[bool] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        reasoning_effort: typing.Optional[str] = OMIT,
        service_tier: typing.Optional[str] = OMIT,
        functions: typing.Optional[typing.Sequence[Function]] = OMIT,
        function_call: typing.Optional[FunctionCall] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[ChatCompletionResponse]:
        """
        Create chat completions

        Parameters
        ----------
        model : typing.Optional[str]

        messages : typing.Optional[typing.Sequence[Message]]

        temperature : typing.Optional[float]

        top_p : typing.Optional[float]

        n : typing.Optional[int]

        stream : typing.Optional[bool]

        stream_options : typing.Optional[StreamOptions]

        stop : typing.Optional[typing.Sequence[str]]

        max_tokens : typing.Optional[int]

        max_completion_tokens : typing.Optional[int]

        presence_penalty : typing.Optional[float]

        frequency_penalty : typing.Optional[float]

        logit_bias : typing.Optional[typing.Dict[str, int]]

        user : typing.Optional[str]

        response_format : typing.Optional[ResponseFormat]

        seed : typing.Optional[int]

        tools : typing.Optional[typing.Sequence[Tool]]

        tool_choice : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]

        parallel_tool_calls : typing.Optional[bool]

        store : typing.Optional[bool]

        metadata : typing.Optional[typing.Dict[str, str]]

        reasoning_effort : typing.Optional[str]

        service_tier : typing.Optional[str]

        functions : typing.Optional[typing.Sequence[Function]]

        function_call : typing.Optional[FunctionCall]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ChatCompletionResponse]
            Chat completions response
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/chat/completions",
            method="POST",
            json={
                "model": model,
                "messages": messages,
                "temperature": temperature,
                "top_p": top_p,
                "n": n,
                "stream": stream,
                "stream_options": convert_and_respect_annotation_metadata(
                    object_=stream_options, annotation=StreamOptions, direction="write"
                ),
                "stop": stop,
                "max_tokens": max_tokens,
                "max_completion_tokens": max_completion_tokens,
                "presence_penalty": presence_penalty,
                "frequency_penalty": frequency_penalty,
                "logit_bias": logit_bias,
                "user": user,
                "response_format": convert_and_respect_annotation_metadata(
                    object_=response_format, annotation=ResponseFormat, direction="write"
                ),
                "seed": seed,
                "tools": convert_and_respect_annotation_metadata(
                    object_=tools, annotation=typing.Sequence[Tool], direction="write"
                ),
                "tool_choice": tool_choice,
                "parallel_tool_calls": parallel_tool_calls,
                "store": store,
                "metadata": metadata,
                "reasoning_effort": reasoning_effort,
                "service_tier": service_tier,
                "functions": convert_and_respect_annotation_metadata(
                    object_=functions, annotation=typing.Sequence[Function], direction="write"
                ),
                "function_call": convert_and_respect_annotation_metadata(
                    object_=function_call, annotation=FunctionCall, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ChatCompletionResponse,
                    parse_obj_as(
                        type_=ChatCompletionResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



from typing import Any, Callable, Optional, Sequence, Union

try:
    import nltk  # type: ignore
    from nltk.corpus import wordnet  # type: ignore
except ImportError:  # pragma: no cover - optional dependency
    nltk = None
    wordnet = None

from opik.exceptions import MetricComputationError
from opik.evaluation.metrics import base_metric, score_result

try:
    from nltk.translate import meteor_score as nltk_meteor_score
except ImportError:  # pragma: no cover - optional dependency
    nltk_meteor_score = None


MeteorFn = Callable[[Sequence[str], str], float]


class METEOR(base_metric.BaseMetric):
    """Computes the METEOR score between output and reference text.

    This implementation wraps ``nltk.translate.meteor_score.meteor_score`` while
    allowing a custom scoring function to be injected (useful for testing).

    References:
      - Banerjee & Lavie, "METEOR: An Automatic Metric for MT Evaluation with Improved
        Correlation with Human Judgments" (ACL Workshop 2005)
        https://aclanthology.org/W05-0909/
      - Hugging Face Evaluate: METEOR metric overview
        https://huggingface.co/spaces/evaluate-metric/meteor

    Args:
        meteor_fn: Optional callable with the same interface as
            ``nltk.translate.meteor_score.meteor_score``. When omitted the
            function from NLTK is used.
        alpha: Precision weight.
        beta: Penalty exponent.
        gamma: Fragmentation penalty weight.
        name: Optional metric name.
        track: Whether Opik should track the metric automatically.
        project_name: Optional project name used when tracking.
    """

    def __init__(
        self,
        meteor_fn: Optional[MeteorFn] = None,
        alpha: float = 0.9,
        beta: float = 3.0,
        gamma: float = 0.5,
        name: str = "meteor_metric",
        track: bool = True,
        project_name: Optional[str] = None,
    ) -> None:
        super().__init__(name=name, track=track, project_name=project_name)

        if meteor_fn is not None:
            self._meteor_fn = meteor_fn
        else:
            if nltk_meteor_score is None:  # pragma: no cover - optional dependency
                raise ImportError(
                    "METEOR metric requires the optional 'nltk' package. Install via"
                    " `pip install nltk` or provide `meteor_fn`."
                )

            if nltk is not None and wordnet is not None:
                try:
                    wordnet.ensure_loaded()  # type: ignore[attr-defined]
                except (
                    LookupError
                ):  # pragma: no cover - download path relies on network access
                    try:
                        nltk.download("wordnet", quiet=True)
                        nltk.download("omw-1.4", quiet=True)
                        wordnet.ensure_loaded()  # type: ignore[attr-defined]
                    except Exception as download_error:
                        raise ImportError(
                            "METEOR metric requires the NLTK corpora 'wordnet' and 'omw-1.4'. "
                            "Install manually via `python -m nltk.downloader wordnet omw-1.4`."
                        ) from download_error

            def _scorer(references: Sequence[str], hypothesis: str) -> float:
                try:
                    return float(
                        nltk_meteor_score.meteor_score(
                            references, hypothesis, alpha=alpha, beta=beta, gamma=gamma
                        )
                    )
                except LookupError as error:
                    raise MetricComputationError(
                        "NLTK resource requirement for METEOR not satisfied. "
                        "Download WordNet via `nltk.download('wordnet')`."
                    ) from error

            self._meteor_fn = _scorer

    def score(
        self,
        output: str,
        reference: Union[str, Sequence[str]],
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        if not output.strip():
            raise MetricComputationError("Candidate is empty (METEOR metric).")
        if isinstance(reference, str):
            references: Sequence[str] = [reference]
        else:
            references = list(reference)
        if not references or any(not ref.strip() for ref in references):
            raise MetricComputationError("Reference is empty (METEOR metric).")

        score = self._meteor_fn(references, output)
        return score_result.ScoreResult(
            value=float(score),
            name=self.name,
            reason=f"METEOR score: {float(score):.4f}",
        )



# This file was auto-generated by Fern from our API Definition.

import contextlib
import datetime as dt
import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.datetime_utils import serialize_datetime
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.not_found_error import NotFoundError
from ..errors.not_implemented_error import NotImplementedError
from ..types.comment import Comment
from ..types.error_info import ErrorInfo
from ..types.error_info_write import ErrorInfoWrite
from ..types.feedback_score_batch_item import FeedbackScoreBatchItem
from ..types.feedback_score_source import FeedbackScoreSource
from ..types.json_list_string import JsonListString
from ..types.json_list_string_write import JsonListStringWrite
from ..types.project_stats_public import ProjectStatsPublic
from ..types.span_filter_public import SpanFilterPublic
from ..types.span_page_public import SpanPagePublic
from ..types.span_public import SpanPublic
from ..types.span_update import SpanUpdate
from ..types.span_update_type import SpanUpdateType
from ..types.span_write import SpanWrite
from ..types.span_write_type import SpanWriteType
from ..types.value_entry import ValueEntry
from .types.find_feedback_score_names_1_request_type import FindFeedbackScoreNames1RequestType
from .types.get_span_stats_request_type import GetSpanStatsRequestType
from .types.get_spans_by_project_request_type import GetSpansByProjectRequestType
from .types.span_search_stream_request_public_type import SpanSearchStreamRequestPublicType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawSpansClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def add_span_comment(
        self,
        id_: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Add span comment

        Parameters
        ----------
        id_ : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id_)}/comments",
            method="POST",
            json={
                "id": id,
                "text": text,
                "created_at": created_at,
                "last_updated_at": last_updated_at,
                "created_by": created_by,
                "last_updated_by": last_updated_by,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def add_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        value: float,
        source: FeedbackScoreSource,
        category_name: typing.Optional[str] = OMIT,
        reason: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        value_by_author: typing.Optional[typing.Dict[str, ValueEntry]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Add span feedback score

        Parameters
        ----------
        id : str

        name : str

        value : float

        source : FeedbackScoreSource

        category_name : typing.Optional[str]

        reason : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        value_by_author : typing.Optional[typing.Dict[str, ValueEntry]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}/feedback-scores",
            method="PUT",
            json={
                "name": name,
                "category_name": category_name,
                "value": value,
                "reason": reason,
                "source": source,
                "created_at": created_at,
                "last_updated_at": last_updated_at,
                "created_by": created_by,
                "last_updated_by": last_updated_by,
                "value_by_author": convert_and_respect_annotation_metadata(
                    object_=value_by_author, annotation=typing.Dict[str, ValueEntry], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_spans(
        self, *, spans: typing.Sequence[SpanWrite], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Create spans

        Parameters
        ----------
        spans : typing.Sequence[SpanWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans/batch",
            method="POST",
            json={
                "spans": convert_and_respect_annotation_metadata(
                    object_=spans, annotation=typing.Sequence[SpanWrite], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def batch_update_spans(
        self,
        *,
        ids: typing.Sequence[str],
        update: SpanUpdate,
        merge_tags: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update multiple spans

        Parameters
        ----------
        ids : typing.Sequence[str]
            List of span IDs to update (max 1000)

        update : SpanUpdate

        merge_tags : typing.Optional[bool]
            If true, merge tags with existing tags instead of replacing them. Default: false

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans/batch",
            method="PATCH",
            json={
                "ids": ids,
                "update": convert_and_respect_annotation_metadata(
                    object_=update, annotation=SpanUpdate, direction="write"
                ),
                "merge_tags": merge_tags,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_spans_by_project(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        project_name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpansByProjectRequestType] = None,
        filters: typing.Optional[str] = None,
        truncate: typing.Optional[bool] = None,
        strip_attachments: typing.Optional[bool] = None,
        sorting: typing.Optional[str] = None,
        exclude: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[SpanPagePublic]:
        """
        Get spans by project_name or project_id and optionally by trace_id and/or type

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpansByProjectRequestType]

        filters : typing.Optional[str]

        truncate : typing.Optional[bool]

        strip_attachments : typing.Optional[bool]

        sorting : typing.Optional[str]

        exclude : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[SpanPagePublic]
            Spans resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans",
            method="GET",
            params={
                "page": page,
                "size": size,
                "project_name": project_name,
                "project_id": project_id,
                "trace_id": trace_id,
                "type": type,
                "filters": filters,
                "truncate": truncate,
                "strip_attachments": strip_attachments,
                "sorting": sorting,
                "exclude": exclude,
                "from_time": serialize_datetime(from_time) if from_time is not None else None,
                "to_time": serialize_datetime(to_time) if to_time is not None else None,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SpanPagePublic,
                    parse_obj_as(
                        type_=SpanPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_span(
        self,
        *,
        start_time: dt.datetime,
        id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanWriteType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListStringWrite] = OMIT,
        output: typing.Optional[JsonListStringWrite] = OMIT,
        metadata: typing.Optional[JsonListStringWrite] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        error_info: typing.Optional[ErrorInfoWrite] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        total_estimated_cost_version: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create span

        Parameters
        ----------
        start_time : dt.datetime

        id : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        trace_id : typing.Optional[str]

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanWriteType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListStringWrite]

        output : typing.Optional[JsonListStringWrite]

        metadata : typing.Optional[JsonListStringWrite]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        usage : typing.Optional[typing.Dict[str, int]]

        error_info : typing.Optional[ErrorInfoWrite]

        last_updated_at : typing.Optional[dt.datetime]

        total_estimated_cost : typing.Optional[float]

        total_estimated_cost_version : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans",
            method="POST",
            json={
                "id": id,
                "project_name": project_name,
                "trace_id": trace_id,
                "parent_span_id": parent_span_id,
                "name": name,
                "type": type,
                "start_time": start_time,
                "end_time": end_time,
                "input": convert_and_respect_annotation_metadata(
                    object_=input, annotation=JsonListStringWrite, direction="write"
                ),
                "output": convert_and_respect_annotation_metadata(
                    object_=output, annotation=JsonListStringWrite, direction="write"
                ),
                "metadata": convert_and_respect_annotation_metadata(
                    object_=metadata, annotation=JsonListStringWrite, direction="write"
                ),
                "model": model,
                "provider": provider,
                "tags": tags,
                "usage": usage,
                "error_info": convert_and_respect_annotation_metadata(
                    object_=error_info, annotation=ErrorInfoWrite, direction="write"
                ),
                "last_updated_at": last_updated_at,
                "total_estimated_cost": total_estimated_cost,
                "total_estimated_cost_version": total_estimated_cost_version,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_span_by_id(
        self,
        id: str,
        *,
        strip_attachments: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[SpanPublic]:
        """
        Get span by id

        Parameters
        ----------
        id : str

        strip_attachments : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[SpanPublic]
            Span resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}",
            method="GET",
            params={
                "strip_attachments": strip_attachments,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SpanPublic,
                    parse_obj_as(
                        type_=SpanPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_span_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete span by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 501:
                raise NotImplementedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_span(
        self,
        id: str,
        *,
        trace_id: str,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanUpdateType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListString] = OMIT,
        output: typing.Optional[JsonListString] = OMIT,
        metadata: typing.Optional[JsonListString] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        error_info: typing.Optional[ErrorInfo] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update span by id

        Parameters
        ----------
        id : str

        trace_id : str

        project_name : typing.Optional[str]
            If null and project_id not specified, Default Project is assumed

        project_id : typing.Optional[str]
            If null and project_name not specified, Default Project is assumed

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanUpdateType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListString]

        output : typing.Optional[JsonListString]

        metadata : typing.Optional[JsonListString]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        usage : typing.Optional[typing.Dict[str, int]]

        total_estimated_cost : typing.Optional[float]

        error_info : typing.Optional[ErrorInfo]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "project_name": project_name,
                "project_id": project_id,
                "trace_id": trace_id,
                "parent_span_id": parent_span_id,
                "name": name,
                "type": type,
                "end_time": end_time,
                "input": convert_and_respect_annotation_metadata(
                    object_=input, annotation=JsonListString, direction="write"
                ),
                "output": convert_and_respect_annotation_metadata(
                    object_=output, annotation=JsonListString, direction="write"
                ),
                "metadata": convert_and_respect_annotation_metadata(
                    object_=metadata, annotation=JsonListString, direction="write"
                ),
                "model": model,
                "provider": provider,
                "tags": tags,
                "usage": usage,
                "total_estimated_cost": total_estimated_cost,
                "error_info": convert_and_respect_annotation_metadata(
                    object_=error_info, annotation=ErrorInfo, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_span_comments(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete span comments

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans/comments/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        author: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Delete span feedback score

        Parameters
        ----------
        id : str

        name : str

        author : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}/feedback-scores/delete",
            method="POST",
            json={
                "name": name,
                "author": author,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def find_feedback_score_names_1(
        self,
        *,
        project_id: typing.Optional[str] = None,
        type: typing.Optional[FindFeedbackScoreNames1RequestType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[typing.List[str]]:
        """
        Find Feedback Score names

        Parameters
        ----------
        project_id : typing.Optional[str]

        type : typing.Optional[FindFeedbackScoreNames1RequestType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[typing.List[str]]
            Feedback Scores resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans/feedback-scores/names",
            method="GET",
            params={
                "project_id": project_id,
                "type": type,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[str],
                    parse_obj_as(
                        type_=typing.List[str],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_span_comment(
        self, comment_id: str, span_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[Comment]:
        """
        Get span comment

        Parameters
        ----------
        comment_id : str

        span_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[Comment]
            Comment resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(span_id)}/comments/{jsonable_encoder(comment_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    Comment,
                    parse_obj_as(
                        type_=Comment,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_span_stats(
        self,
        *,
        project_id: typing.Optional[str] = None,
        project_name: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpanStatsRequestType] = None,
        filters: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[ProjectStatsPublic]:
        """
        Get span stats

        Parameters
        ----------
        project_id : typing.Optional[str]

        project_name : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpanStatsRequestType]

        filters : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProjectStatsPublic]
            Span stats resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans/stats",
            method="GET",
            params={
                "project_id": project_id,
                "project_name": project_name,
                "trace_id": trace_id,
                "type": type,
                "filters": filters,
                "from_time": serialize_datetime(from_time) if from_time is not None else None,
                "to_time": serialize_datetime(to_time) if to_time is not None else None,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectStatsPublic,
                    parse_obj_as(
                        type_=ProjectStatsPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def score_batch_of_spans(
        self,
        *,
        scores: typing.Sequence[FeedbackScoreBatchItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Batch feedback scoring for spans

        Parameters
        ----------
        scores : typing.Sequence[FeedbackScoreBatchItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/spans/feedback-scores",
            method="PUT",
            json={
                "scores": convert_and_respect_annotation_metadata(
                    object_=scores, annotation=typing.Sequence[FeedbackScoreBatchItem], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    @contextlib.contextmanager
    def search_spans(
        self,
        *,
        trace_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanSearchStreamRequestPublicType] = OMIT,
        filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        truncate: typing.Optional[bool] = OMIT,
        from_time: typing.Optional[dt.datetime] = OMIT,
        to_time: typing.Optional[dt.datetime] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[HttpResponse[typing.Iterator[bytes]]]:
        """
        Search spans

        Parameters
        ----------
        trace_id : typing.Optional[str]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        type : typing.Optional[SpanSearchStreamRequestPublicType]

        filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        limit : typing.Optional[int]
            Max number of spans to be streamed

        last_retrieved_id : typing.Optional[str]

        truncate : typing.Optional[bool]
            Truncate image included in either input, output or metadata

        from_time : typing.Optional[dt.datetime]
            Filter spans created from this time (ISO-8601 format).

        to_time : typing.Optional[dt.datetime]
            Filter spans created up to this time (ISO-8601 format). If not provided, defaults to current time. Must be after 'from_time'.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.Iterator[HttpResponse[typing.Iterator[bytes]]]
            Spans stream or error during process
        """
        with self._client_wrapper.httpx_client.stream(
            "v1/private/spans/search",
            method="POST",
            json={
                "trace_id": trace_id,
                "project_name": project_name,
                "project_id": project_id,
                "type": type,
                "filters": convert_and_respect_annotation_metadata(
                    object_=filters, annotation=typing.Sequence[SpanFilterPublic], direction="write"
                ),
                "limit": limit,
                "last_retrieved_id": last_retrieved_id,
                "truncate": truncate,
                "from_time": from_time,
                "to_time": to_time,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        ) as _response:

            def stream() -> HttpResponse[typing.Iterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return HttpResponse(
                            response=_response, data=(_chunk for _chunk in _response.iter_bytes(chunk_size=_chunk_size))
                        )
                    _response.read()
                    if _response.status_code == 400:
                        raise BadRequestError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield stream()

    def update_span_comment(
        self,
        comment_id: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update span comment by id

        Parameters
        ----------
        comment_id : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/spans/comments/{jsonable_encoder(comment_id)}",
            method="PATCH",
            json={
                "id": id,
                "text": text,
                "created_at": created_at,
                "last_updated_at": last_updated_at,
                "created_by": created_by,
                "last_updated_by": last_updated_by,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawSpansClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def add_span_comment(
        self,
        id_: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Add span comment

        Parameters
        ----------
        id_ : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id_)}/comments",
            method="POST",
            json={
                "id": id,
                "text": text,
                "created_at": created_at,
                "last_updated_at": last_updated_at,
                "created_by": created_by,
                "last_updated_by": last_updated_by,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def add_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        value: float,
        source: FeedbackScoreSource,
        category_name: typing.Optional[str] = OMIT,
        reason: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        value_by_author: typing.Optional[typing.Dict[str, ValueEntry]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Add span feedback score

        Parameters
        ----------
        id : str

        name : str

        value : float

        source : FeedbackScoreSource

        category_name : typing.Optional[str]

        reason : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        value_by_author : typing.Optional[typing.Dict[str, ValueEntry]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}/feedback-scores",
            method="PUT",
            json={
                "name": name,
                "category_name": category_name,
                "value": value,
                "reason": reason,
                "source": source,
                "created_at": created_at,
                "last_updated_at": last_updated_at,
                "created_by": created_by,
                "last_updated_by": last_updated_by,
                "value_by_author": convert_and_respect_annotation_metadata(
                    object_=value_by_author, annotation=typing.Dict[str, ValueEntry], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_spans(
        self, *, spans: typing.Sequence[SpanWrite], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Create spans

        Parameters
        ----------
        spans : typing.Sequence[SpanWrite]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans/batch",
            method="POST",
            json={
                "spans": convert_and_respect_annotation_metadata(
                    object_=spans, annotation=typing.Sequence[SpanWrite], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def batch_update_spans(
        self,
        *,
        ids: typing.Sequence[str],
        update: SpanUpdate,
        merge_tags: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update multiple spans

        Parameters
        ----------
        ids : typing.Sequence[str]
            List of span IDs to update (max 1000)

        update : SpanUpdate

        merge_tags : typing.Optional[bool]
            If true, merge tags with existing tags instead of replacing them. Default: false

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans/batch",
            method="PATCH",
            json={
                "ids": ids,
                "update": convert_and_respect_annotation_metadata(
                    object_=update, annotation=SpanUpdate, direction="write"
                ),
                "merge_tags": merge_tags,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_spans_by_project(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        project_name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpansByProjectRequestType] = None,
        filters: typing.Optional[str] = None,
        truncate: typing.Optional[bool] = None,
        strip_attachments: typing.Optional[bool] = None,
        sorting: typing.Optional[str] = None,
        exclude: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[SpanPagePublic]:
        """
        Get spans by project_name or project_id and optionally by trace_id and/or type

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpansByProjectRequestType]

        filters : typing.Optional[str]

        truncate : typing.Optional[bool]

        strip_attachments : typing.Optional[bool]

        sorting : typing.Optional[str]

        exclude : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[SpanPagePublic]
            Spans resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans",
            method="GET",
            params={
                "page": page,
                "size": size,
                "project_name": project_name,
                "project_id": project_id,
                "trace_id": trace_id,
                "type": type,
                "filters": filters,
                "truncate": truncate,
                "strip_attachments": strip_attachments,
                "sorting": sorting,
                "exclude": exclude,
                "from_time": serialize_datetime(from_time) if from_time is not None else None,
                "to_time": serialize_datetime(to_time) if to_time is not None else None,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SpanPagePublic,
                    parse_obj_as(
                        type_=SpanPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_span(
        self,
        *,
        start_time: dt.datetime,
        id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanWriteType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListStringWrite] = OMIT,
        output: typing.Optional[JsonListStringWrite] = OMIT,
        metadata: typing.Optional[JsonListStringWrite] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        error_info: typing.Optional[ErrorInfoWrite] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        total_estimated_cost_version: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create span

        Parameters
        ----------
        start_time : dt.datetime

        id : typing.Optional[str]

        project_name : typing.Optional[str]
            If null, the default project is used

        trace_id : typing.Optional[str]

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanWriteType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListStringWrite]

        output : typing.Optional[JsonListStringWrite]

        metadata : typing.Optional[JsonListStringWrite]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        usage : typing.Optional[typing.Dict[str, int]]

        error_info : typing.Optional[ErrorInfoWrite]

        last_updated_at : typing.Optional[dt.datetime]

        total_estimated_cost : typing.Optional[float]

        total_estimated_cost_version : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans",
            method="POST",
            json={
                "id": id,
                "project_name": project_name,
                "trace_id": trace_id,
                "parent_span_id": parent_span_id,
                "name": name,
                "type": type,
                "start_time": start_time,
                "end_time": end_time,
                "input": convert_and_respect_annotation_metadata(
                    object_=input, annotation=JsonListStringWrite, direction="write"
                ),
                "output": convert_and_respect_annotation_metadata(
                    object_=output, annotation=JsonListStringWrite, direction="write"
                ),
                "metadata": convert_and_respect_annotation_metadata(
                    object_=metadata, annotation=JsonListStringWrite, direction="write"
                ),
                "model": model,
                "provider": provider,
                "tags": tags,
                "usage": usage,
                "error_info": convert_and_respect_annotation_metadata(
                    object_=error_info, annotation=ErrorInfoWrite, direction="write"
                ),
                "last_updated_at": last_updated_at,
                "total_estimated_cost": total_estimated_cost,
                "total_estimated_cost_version": total_estimated_cost_version,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_span_by_id(
        self,
        id: str,
        *,
        strip_attachments: typing.Optional[bool] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[SpanPublic]:
        """
        Get span by id

        Parameters
        ----------
        id : str

        strip_attachments : typing.Optional[bool]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[SpanPublic]
            Span resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}",
            method="GET",
            params={
                "strip_attachments": strip_attachments,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    SpanPublic,
                    parse_obj_as(
                        type_=SpanPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_span_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete span by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 501:
                raise NotImplementedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_span(
        self,
        id: str,
        *,
        trace_id: str,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        parent_span_id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanUpdateType] = OMIT,
        end_time: typing.Optional[dt.datetime] = OMIT,
        input: typing.Optional[JsonListString] = OMIT,
        output: typing.Optional[JsonListString] = OMIT,
        metadata: typing.Optional[JsonListString] = OMIT,
        model: typing.Optional[str] = OMIT,
        provider: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        usage: typing.Optional[typing.Dict[str, int]] = OMIT,
        total_estimated_cost: typing.Optional[float] = OMIT,
        error_info: typing.Optional[ErrorInfo] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update span by id

        Parameters
        ----------
        id : str

        trace_id : str

        project_name : typing.Optional[str]
            If null and project_id not specified, Default Project is assumed

        project_id : typing.Optional[str]
            If null and project_name not specified, Default Project is assumed

        parent_span_id : typing.Optional[str]

        name : typing.Optional[str]

        type : typing.Optional[SpanUpdateType]

        end_time : typing.Optional[dt.datetime]

        input : typing.Optional[JsonListString]

        output : typing.Optional[JsonListString]

        metadata : typing.Optional[JsonListString]

        model : typing.Optional[str]

        provider : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        usage : typing.Optional[typing.Dict[str, int]]

        total_estimated_cost : typing.Optional[float]

        error_info : typing.Optional[ErrorInfo]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "project_name": project_name,
                "project_id": project_id,
                "trace_id": trace_id,
                "parent_span_id": parent_span_id,
                "name": name,
                "type": type,
                "end_time": end_time,
                "input": convert_and_respect_annotation_metadata(
                    object_=input, annotation=JsonListString, direction="write"
                ),
                "output": convert_and_respect_annotation_metadata(
                    object_=output, annotation=JsonListString, direction="write"
                ),
                "metadata": convert_and_respect_annotation_metadata(
                    object_=metadata, annotation=JsonListString, direction="write"
                ),
                "model": model,
                "provider": provider,
                "tags": tags,
                "usage": usage,
                "total_estimated_cost": total_estimated_cost,
                "error_info": convert_and_respect_annotation_metadata(
                    object_=error_info, annotation=ErrorInfo, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_span_comments(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete span comments

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans/comments/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_span_feedback_score(
        self,
        id: str,
        *,
        name: str,
        author: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Delete span feedback score

        Parameters
        ----------
        id : str

        name : str

        author : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(id)}/feedback-scores/delete",
            method="POST",
            json={
                "name": name,
                "author": author,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def find_feedback_score_names_1(
        self,
        *,
        project_id: typing.Optional[str] = None,
        type: typing.Optional[FindFeedbackScoreNames1RequestType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[typing.List[str]]:
        """
        Find Feedback Score names

        Parameters
        ----------
        project_id : typing.Optional[str]

        type : typing.Optional[FindFeedbackScoreNames1RequestType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[typing.List[str]]
            Feedback Scores resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans/feedback-scores/names",
            method="GET",
            params={
                "project_id": project_id,
                "type": type,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[str],
                    parse_obj_as(
                        type_=typing.List[str],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_span_comment(
        self, comment_id: str, span_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[Comment]:
        """
        Get span comment

        Parameters
        ----------
        comment_id : str

        span_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[Comment]
            Comment resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/{jsonable_encoder(span_id)}/comments/{jsonable_encoder(comment_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    Comment,
                    parse_obj_as(
                        type_=Comment,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_span_stats(
        self,
        *,
        project_id: typing.Optional[str] = None,
        project_name: typing.Optional[str] = None,
        trace_id: typing.Optional[str] = None,
        type: typing.Optional[GetSpanStatsRequestType] = None,
        filters: typing.Optional[str] = None,
        from_time: typing.Optional[dt.datetime] = None,
        to_time: typing.Optional[dt.datetime] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[ProjectStatsPublic]:
        """
        Get span stats

        Parameters
        ----------
        project_id : typing.Optional[str]

        project_name : typing.Optional[str]

        trace_id : typing.Optional[str]

        type : typing.Optional[GetSpanStatsRequestType]

        filters : typing.Optional[str]

        from_time : typing.Optional[dt.datetime]

        to_time : typing.Optional[dt.datetime]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProjectStatsPublic]
            Span stats resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans/stats",
            method="GET",
            params={
                "project_id": project_id,
                "project_name": project_name,
                "trace_id": trace_id,
                "type": type,
                "filters": filters,
                "from_time": serialize_datetime(from_time) if from_time is not None else None,
                "to_time": serialize_datetime(to_time) if to_time is not None else None,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectStatsPublic,
                    parse_obj_as(
                        type_=ProjectStatsPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def score_batch_of_spans(
        self,
        *,
        scores: typing.Sequence[FeedbackScoreBatchItem],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Batch feedback scoring for spans

        Parameters
        ----------
        scores : typing.Sequence[FeedbackScoreBatchItem]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/spans/feedback-scores",
            method="PUT",
            json={
                "scores": convert_and_respect_annotation_metadata(
                    object_=scores, annotation=typing.Sequence[FeedbackScoreBatchItem], direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    @contextlib.asynccontextmanager
    async def search_spans(
        self,
        *,
        trace_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        type: typing.Optional[SpanSearchStreamRequestPublicType] = OMIT,
        filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        limit: typing.Optional[int] = OMIT,
        last_retrieved_id: typing.Optional[str] = OMIT,
        truncate: typing.Optional[bool] = OMIT,
        from_time: typing.Optional[dt.datetime] = OMIT,
        to_time: typing.Optional[dt.datetime] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]:
        """
        Search spans

        Parameters
        ----------
        trace_id : typing.Optional[str]

        project_name : typing.Optional[str]

        project_id : typing.Optional[str]

        type : typing.Optional[SpanSearchStreamRequestPublicType]

        filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        limit : typing.Optional[int]
            Max number of spans to be streamed

        last_retrieved_id : typing.Optional[str]

        truncate : typing.Optional[bool]
            Truncate image included in either input, output or metadata

        from_time : typing.Optional[dt.datetime]
            Filter spans created from this time (ISO-8601 format).

        to_time : typing.Optional[dt.datetime]
            Filter spans created up to this time (ISO-8601 format). If not provided, defaults to current time. Must be after 'from_time'.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration. You can pass in configuration such as `chunk_size`, and more to customize the request and response.

        Returns
        -------
        typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[bytes]]]
            Spans stream or error during process
        """
        async with self._client_wrapper.httpx_client.stream(
            "v1/private/spans/search",
            method="POST",
            json={
                "trace_id": trace_id,
                "project_name": project_name,
                "project_id": project_id,
                "type": type,
                "filters": convert_and_respect_annotation_metadata(
                    object_=filters, annotation=typing.Sequence[SpanFilterPublic], direction="write"
                ),
                "limit": limit,
                "last_retrieved_id": last_retrieved_id,
                "truncate": truncate,
                "from_time": from_time,
                "to_time": to_time,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        ) as _response:

            async def stream() -> AsyncHttpResponse[typing.AsyncIterator[bytes]]:
                try:
                    if 200 <= _response.status_code < 300:
                        _chunk_size = request_options.get("chunk_size", None) if request_options is not None else None
                        return AsyncHttpResponse(
                            response=_response,
                            data=(_chunk async for _chunk in _response.aiter_bytes(chunk_size=_chunk_size)),
                        )
                    await _response.aread()
                    if _response.status_code == 400:
                        raise BadRequestError(
                            headers=dict(_response.headers),
                            body=typing.cast(
                                typing.Optional[typing.Any],
                                parse_obj_as(
                                    type_=typing.Optional[typing.Any],  # type: ignore
                                    object_=_response.json(),
                                ),
                            ),
                        )
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(
                        status_code=_response.status_code, headers=dict(_response.headers), body=_response.text
                    )
                raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

            yield await stream()

    async def update_span_comment(
        self,
        comment_id: str,
        *,
        text: str,
        id: typing.Optional[str] = OMIT,
        created_at: typing.Optional[dt.datetime] = OMIT,
        last_updated_at: typing.Optional[dt.datetime] = OMIT,
        created_by: typing.Optional[str] = OMIT,
        last_updated_by: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update span comment by id

        Parameters
        ----------
        comment_id : str

        text : str

        id : typing.Optional[str]

        created_at : typing.Optional[dt.datetime]

        last_updated_at : typing.Optional[dt.datetime]

        created_by : typing.Optional[str]

        last_updated_by : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/spans/comments/{jsonable_encoder(comment_id)}",
            method="PATCH",
            json={
                "id": id,
                "text": text,
                "created_at": created_at,
                "last_updated_at": last_updated_at,
                "created_by": created_by,
                "last_updated_by": last_updated_by,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



"""Utility helpers shared across LiteLLM models."""

from __future__ import annotations

from typing import Any, Callable, Dict, Set


def normalise_choice(choice: Any) -> Dict[str, Any]:
    """Produce a dict view of a LiteLLM choice regardless of response type.

    LiteLLM may return raw dicts, Pydantic models, or dataclasses. Normalising to a
    dict here keeps downstream parsing logic consistent and backwards compatible with
    older client versions.
    """

    if isinstance(choice, dict):
        return choice
    if hasattr(choice, "model_dump") and callable(choice.model_dump):
        try:
            return choice.model_dump()
        except TypeError:
            pass
    normalised: Dict[str, Any] = {}
    message = getattr(choice, "message", None)
    if message is not None:
        normalised["message"] = message
    logprobs = getattr(choice, "logprobs", None)
    if logprobs is not None:
        normalised["logprobs"] = logprobs
    return normalised


def apply_model_specific_filters(
    model_name: str,
    params: Dict[str, Any],
    already_warned: Set[str],
    warn: Callable[[str, Any], None],
) -> None:
    """Adjust/drop params for specific model families before calling LiteLLM.

    Currently handles:
    - GPT-5: only honours temperature=1 and does not return log probabilities.
    - DashScope Qwen: enforces constraints for logprobs / top_logprobs
    """
    if model_name.startswith("gpt-5"):
        _apply_gpt5_filters(params, already_warned, warn)
        return

    if model_name.startswith("dashscope/"):
        _apply_qwen_dashscope_filters(params, already_warned, warn)
        return


def _apply_gpt5_filters(
    params: Dict[str, Any],
    already_warned: Set[str],
    warn: Callable[[str, Any], None],
) -> None:
    """Apply GPT-5 specific parameter filters.

    Only honours temperature=1 and does not return log probabilities.
    Removing those eagerly avoids provider errors while the callback surfaces a
    one-time warning to the caller.
    """

    unsupported: list[tuple[str, Any]] = []

    if "temperature" in params:
        value = params["temperature"]
        try:
            numeric_value = float(value)
        except (TypeError, ValueError):
            numeric_value = None
        if numeric_value is None or abs(numeric_value - 1.0) > 1e-6:
            unsupported.append(("temperature", value))

    for param in ("logprobs", "top_logprobs"):
        if param in params:
            unsupported.append((param, params[param]))

    _drop_unsupported_params_with_warning(
        params,
        unsupported,
        already_warned,
        warn,
    )


def _apply_qwen_dashscope_filters(
    params: Dict[str, Any],
    already_warned: Set[str],
    warn: Callable[[str, Any], None],
) -> None:
    """Apply Qwen/DashScope specific parameter filters.

    Does not return log probabilities.
    """

    unsupported: list[tuple[str, Any]] = []

    for param in ("logprobs", "top_logprobs"):
        if param in params:
            unsupported.append((param, params[param]))

    _drop_unsupported_params_with_warning(
        params,
        unsupported,
        already_warned,
        warn,
    )


def _drop_unsupported_params_with_warning(
    params: Dict[str, Any],
    unsupported_params: list[tuple[str, Any]],
    already_warned: Set[str],
    warn: Callable[[str, Any], None],
) -> None:
    """Remove unsupported params and emit warnings once per param name."""
    for param, value in unsupported_params:
        params.pop(param, None)
        if param in already_warned:
            continue
        warn(param, value)
        already_warned.add(param)



from typing import Union, Optional, List, Any
import pydantic

from opik.evaluation.models import base_model, models_factory
from opik.evaluation.metrics import score_result, base_metric

from . import template, parser


class HallucinationResponseFormat(pydantic.BaseModel):
    score: float
    reason: List[str]


class Hallucination(base_metric.BaseMetric):
    """
    A metric that evaluates whether an LLM's output contains hallucinations based on given input and context.

    This metric uses another LLM to judge if the output is factual or contains hallucinations.
    It returns a score of 1.0 if hallucination is detected, and 0.0 otherwise.

    Args:
        model: The LLM to use for evaluation. Can be a string (model name) or an `opik.evaluation.models.OpikBaseModel` subclass instance.
            `opik.evaluation.models.LiteLLMChatModel` is used by default.
        name: The name of the metric.
        few_shot_examples: A list of few-shot examples to use for hallucination detection.  If None, default examples will be used.
        track: Whether to track the metric. Defaults to True.
        project_name: Optional project name to track the metric in for the cases when
            there are no parent span/trace to inherit project name from.
        seed: Optional seed value for reproducible model generation. If provided, this seed will be passed to the model for deterministic outputs.
        temperature: Optional temperature value for model generation. If provided, this temperature will be passed to the model. If not provided, the model's default temperature will be used.

    Example:
        >>> from opik.evaluation.metrics import Hallucination
        >>> hallucination_metric = Hallucination()
        >>> result = hallucination_metric.score(
        ...     input="What is the capital of France?",
        ...     output="The capital of France is London.",
        ...     context=["The capital of France is Paris."]
        ... )
        >>> print(result.value)
        1.0
        >>> print(result.reason)
        The answer provided states that the capital of France is London, which contradicts the fact stated in the context that the capital of France is Paris.
    """

    def __init__(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]] = None,
        name: str = "hallucination_metric",
        few_shot_examples: Optional[List[template.FewShotExampleHallucination]] = None,
        track: bool = True,
        project_name: Optional[str] = None,
        seed: Optional[int] = None,
        temperature: Optional[float] = None,
    ):
        super().__init__(name=name, track=track, project_name=project_name)
        self._seed = seed
        self._init_model(model, temperature=temperature)
        self.few_shot_examples = few_shot_examples

    def _init_model(
        self,
        model: Optional[Union[str, base_model.OpikBaseModel]],
        temperature: Optional[float],
    ) -> None:
        if isinstance(model, base_model.OpikBaseModel):
            self._model = model
        else:
            model_kwargs = {}
            if temperature is not None:
                model_kwargs["temperature"] = temperature
            if self._seed is not None:
                model_kwargs["seed"] = self._seed

            self._model = models_factory.get(model_name=model, **model_kwargs)

    def score(
        self,
        input: str,
        output: str,
        context: Optional[List[str]] = None,
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        """
        Calculate the hallucination score for the given input, output, and optional context field.

        Args:
            input: The original input/question.
            output: The LLM's output to evaluate.
            context: A list of context strings. If not provided, the presence of hallucinations will be evaluated based on the output only.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with a value of 1.0 if hallucination
                is detected, 0.0 otherwise, along with the reason for the verdict.
        """
        llm_query = template.generate_query(
            input=input,
            output=output,
            context=context,
            few_shot_examples=self.few_shot_examples,
        )
        model_output = self._model.generate_string(
            input=llm_query, response_format=HallucinationResponseFormat
        )

        return parser.parse_model_output(content=model_output, name=self.name)

    async def ascore(
        self,
        input: str,
        output: str,
        context: Optional[List[str]] = None,
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        """
        Asynchronously calculate the hallucination score for the given input, output, and optional context field.

        Args:
            input: The original input/question.
            output: The LLM's output to evaluate.
            context: A list of context strings. If not provided, the presence of hallucinations will be evaluated based on the output only.
            **ignored_kwargs: Additional keyword arguments that are ignored.

        Returns:
            score_result.ScoreResult: A ScoreResult object with a value of 1.0 if hallucination
                is detected, 0.0 otherwise, along with the reason for the verdict.
        """
        llm_query = template.generate_query(
            input=input,
            output=output,
            context=context,
            few_shot_examples=self.few_shot_examples,
        )
        model_output = await self._model.agenerate_string(
            input=llm_query, response_format=HallucinationResponseFormat
        )

        return parser.parse_model_output(content=model_output, name=self.name)



from typing import Dict, Any, Optional
from . import base_original_provider_usage


class BedrockUsage(base_original_provider_usage.BaseOriginalProviderUsage):
    """Bedrock calls token usage data. Updated 04.08.2025"""

    inputTokens: int
    """The number of input tokens which were used."""

    outputTokens: int
    """The number of output tokens which were used."""

    cacheReadInputTokens: Optional[int] = None
    """The number of input tokens which were read from the cache."""

    cacheWriteInputTokens: Optional[int] = None
    """The number of input tokens which were written to the cache."""

    def to_backend_compatible_flat_dict(self, parent_key_prefix: str) -> Dict[str, int]:
        return super().to_backend_compatible_flat_dict(parent_key_prefix)

    @classmethod
    def from_original_usage_dict(cls, usage: Dict[str, Any]) -> "BedrockUsage":
        return cls(**usage)



# type: ignore

# The content of this module is copied from comet-ml.

import collections
import re
from functools import wraps
from typing import Callable, Dict, List, Optional, Tuple, Union

VersionPart = Union[int, Optional[str]]
VersionTuple = Tuple[
    int,
    int,
    int,
    Optional[Union[str, int]],
    Optional[Union[str, int]],
    Optional[Union[str, int]],
]

ComparableVersion = Union[
    "SemanticVersion", Dict[str, VersionPart], List[VersionPart], VersionTuple, str
]
Comparator = Callable[["SemanticVersion", ComparableVersion], bool]


def _cmp(a, b) -> int:
    return (a > b) - (a < b)


def _comparator(operator: Comparator) -> Comparator:
    @wraps(operator)
    def wrapper(self: "SemanticVersion", other: ComparableVersion) -> bool:
        comparable_types = (
            SemanticVersion,
            dict,
            tuple,
            list,
            str,
        )
        if not isinstance(other, comparable_types):
            return NotImplemented
        return operator(self, other)

    return wrapper


class SemanticVersion:
    # Based on regex from https://semver.org
    # Regex template for a semver version
    _SEMVER_REGEX_TEMPLATE = r"""
                ^
                (?P<major>0|[1-9]\d*)
                (?:\.(?P<minor>0|[1-9]\d*)
                    (?:-(?P<feature_branch>0|[1-9]\d*|[a-zA-Z-_][0-9a-zA-Z-_]*))?
                    (?:\.(?P<patch>0|[1-9]\d*)){opt_patch}
                ){opt_minor}
                (?:-(?P<pre_release>
                    (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                    (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
                ))?
                (?:\+(?P<build>
                    [0-9a-zA-Z-]+
                    (?:\.[0-9a-zA-Z-]+)*
                ))?
                $
            """
    # Regex for a semver version
    _SEMVER_REGEX = re.compile(
        _SEMVER_REGEX_TEMPLATE.format(opt_patch="", opt_minor=""),
        re.VERBOSE,
    )
    # Regex for a semver version that might be shorter
    _SEMVER_REGEX_OPTIONAL_MINOR_AND_PATCH = re.compile(
        _SEMVER_REGEX_TEMPLATE.format(opt_patch="?", opt_minor="?"),
        re.VERBOSE,
    )

    def __init__(
        self,
        major: int,
        minor: int = 0,
        patch: int = 0,
        feature_branch: Optional[Union[str, int]] = None,
        pre_release: Optional[Union[str, int]] = None,
        build: Optional[Union[str, int]] = None,
    ):
        self._major = major
        self._minor = minor
        self._patch = patch
        self._feature_branch = None if feature_branch is None else str(feature_branch)
        self._pre_release = None if pre_release is None else str(pre_release)
        self._build = None if build is None else str(build)

    @property
    def major(self) -> int:
        return self._major

    @property
    def minor(self) -> int:
        return self._minor

    @property
    def patch(self) -> int:
        return self._patch

    @property
    def pre_release(self) -> Optional[str]:
        return self._pre_release

    @property
    def build(self) -> Optional[str]:
        return self._build

    @property
    def feature_branch(self) -> Optional[str]:
        return self._feature_branch

    def to_tuple(self) -> VersionTuple:
        return (
            self.major,
            self.minor,
            self.patch,
            self.feature_branch,
            self.pre_release,
            self.build,
        )

    def to_dict(self) -> collections.OrderedDict:
        return collections.OrderedDict(
            (
                ("major", self._major),
                ("minor", self._minor),
                ("feature_branch", self._feature_branch),
                ("patch", self._patch),
                ("pre_release", self._pre_release),
                ("build", self._build),
            )
        )

    def compare(self, other: ComparableVersion) -> int:
        """
        Compare self with other version.

        :param other: another version
        :return: The return value is negative if self < other,
             zero if self == other and strictly positive if self > other
        """
        cls = type(self)
        if isinstance(other, str):
            other = cls.parse(other)
        elif isinstance(other, dict):
            other = cls(**other)
        elif isinstance(other, (tuple, list)):
            other = cls(*other)
        elif not isinstance(other, cls):
            raise TypeError(
                "Wrong type. Expected str, bytes, dict, tuple, list, or %r instance, but got %r"
                % (cls.__name__, type(other))
            )

        v1 = self.to_tuple()[:3]
        v2 = other.to_tuple()[:3]
        return _cmp(v1, v2)

    @_comparator
    def __eq__(self, other: ComparableVersion) -> bool:  # type: ignore
        return self.compare(other) == 0

    @_comparator
    def __ne__(self, other: ComparableVersion) -> bool:  # type: ignore
        return self.compare(other) != 0

    @_comparator
    def __lt__(self, other: ComparableVersion) -> bool:
        return self.compare(other) < 0

    @_comparator
    def __le__(self, other: ComparableVersion) -> bool:
        return self.compare(other) <= 0

    @_comparator
    def __gt__(self, other: ComparableVersion) -> bool:
        return self.compare(other) > 0

    @_comparator
    def __ge__(self, other: ComparableVersion) -> bool:
        return self.compare(other) >= 0

    def __repr__(self) -> str:
        s = ", ".join("%s=%r" % (key, val) for key, val in self.to_dict().items())
        return "%s(%s)" % (type(self).__name__, s)

    def __str__(self) -> str:
        version = "%d.%d" % (self.major, self.minor)
        if self._feature_branch:
            version += "-%s" % self._feature_branch

        version += ".%d" % self.patch

        if self.pre_release:
            version += "-%s" % self.pre_release
        if self.build:
            version += "+%s" % self.build
        return version

    @classmethod
    def parse(
        cls, version: str, optional_minor_and_patch: bool = False
    ) -> "SemanticVersion":
        if not isinstance(version, str):
            raise TypeError("wrong version string type %r" % type(version))

        if optional_minor_and_patch:
            match = cls._SEMVER_REGEX_OPTIONAL_MINOR_AND_PATCH.match(version)
        else:
            match = cls._SEMVER_REGEX.match(version)
        if match is None:
            raise ValueError("%r is not valid SemVer string" % version)

        version_parts = match.groupdict()
        if not version_parts["minor"]:
            version_parts["minor"] = 0
        if not version_parts["patch"]:
            version_parts["patch"] = 0

        major = int(version_parts["major"])
        minor = int(version_parts["minor"])
        patch = int(version_parts["patch"])
        feature_branch = version_parts.get("feature_branch", None)
        pre_release = version_parts.get("pre_release", None)
        build = version_parts.get("build", None)

        return cls(
            major=major,
            minor=minor,
            patch=patch,
            feature_branch=feature_branch,
            pre_release=pre_release,
            build=build,
        )



import abc
import collections
import datetime
import logging
import threading
from typing import List, Dict, Union, Optional, Any, Type, Callable

from opik import dict_utils
from opik.rest_api.types import span_write, trace_write
from opik.types import ErrorInfoDict, SpanType
from . import models
from .. import messages
from ..processors import message_processors


LOGGER = logging.getLogger(__name__)


class EmulatorMessageProcessor(message_processors.BaseMessageProcessor, abc.ABC):
    """
    This class acts as a stand-in for the actual backend.

    The real message processor uses data from messages passed
    to the process method to send information to the backend.
    In contrast, the emulator does not send any requests — instead,
    it collects the data from incoming messages and stores it
    in its attributes.

    Beyond simple data storage, it also constructs complete trace
    and span trees from the received messages. These trees are defined
    by model classes and their subclasses located in models. Concrete
    model instances are created in the abstract create_* methods,
    which must be implemented by subclasses.

    Note: If a new message type is added to the Opik SDK, this
    class must be updated accordingly.

    Args:
        active: Flag indicating whether the emulator is active.
        merge_duplicates: Flag indicating whether duplicates (traces or spans) should
            be merged to retain only unique and updated ones.
    """

    def __init__(self, active: bool, merge_duplicates: bool) -> None:
        self.merge_duplicates = merge_duplicates
        self._active = active

        self._register_handlers()

        self._rlock = threading.RLock()

        self.reset()

    def reset(self) -> None:
        """
        Resets the internal state of the instance to its initial configuration.

        This method clears and re-initializes all internal attributes related to
        processed messages, traces, spans, and feedback scores. It ensures that
        the instance can start fresh, as if it was just created, without retaining
        any previous data.
        """
        with self._rlock:
            self._trace_trees: List[models.TraceModel] = []

            self._traces_to_spans_mapping: Dict[str, List[str]] = (
                collections.defaultdict(list)
            )
            # the same as _trace_trees but without a trace. Useful for distributed tracing.
            self._span_trees: List[models.SpanModel] = []
            self._trace_observations: Dict[str, models.TraceModel] = {}
            self._span_observations: Dict[str, models.SpanModel] = {}

            self._span_to_parent_span: Dict[str, Optional[str]] = {}
            self._span_to_trace: Dict[str, Optional[str]] = {}
            self._trace_to_feedback_scores: Dict[
                str, List[models.FeedbackScoreModel]
            ] = collections.defaultdict(list)
            self._span_to_feedback_scores: Dict[
                str, List[models.FeedbackScoreModel]
            ] = collections.defaultdict(list)
            self._experiment_items: List[models.ExperimentItemModel] = []

    def is_active(self) -> bool:
        with self._rlock:
            return self._active

    def set_active(self, active: bool) -> None:
        with self._rlock:
            self._active = active

    @property
    def trace_trees(self) -> List[models.TraceModel]:
        """
        Builds a list of trace trees based on the data from the processed messages.
        Before processing traces, builds span_trees
        """
        with self._rlock:
            # call to connect all spans
            self._build_spans_tree()

            for span_id, trace_id in self._span_to_trace.items():
                if trace_id is None:
                    continue

                trace = self._trace_observations[trace_id]
                if self._span_to_parent_span[
                    span_id
                ] is None and not _observation_already_stored(span_id, trace.spans):
                    span = self._span_observations[span_id]
                    trace.spans.append(span)
                    trace.spans.sort(key=lambda x: x.start_time)

            for trace in self._trace_trees:
                trace.feedback_scores = self._trace_to_feedback_scores[trace.id]

            self._trace_trees.sort(key=lambda x: x.start_time)
            return self._trace_trees

    def _save_trace(self, trace: models.TraceModel) -> None:
        if self.merge_duplicates:
            # merge traces with the same id to keep only the latest one
            if trace.id in self._trace_observations:
                existing_trace: models.TraceModel = self._trace_observations[trace.id]
                if trace.end_time is not None:
                    if (
                        existing_trace.end_time is None
                        or trace.end_time > existing_trace.end_time
                    ):
                        # remove the current trace from the list
                        self._trace_trees.remove(existing_trace)

        self._trace_trees.append(trace)
        self._trace_observations[trace.id] = trace

    def _save_span(
        self, span: models.SpanModel, trace_id: str, parent_span_id: Optional[str]
    ) -> None:
        if self.merge_duplicates:
            # merge spans with the same id to keep only the latest one
            if span.id in self._span_observations:
                existing_span = self._span_observations[span.id]
                if span.end_time is not None:
                    if (
                        existing_span.end_time is None
                        or span.end_time > existing_span.end_time
                    ):
                        self._span_trees.remove(existing_span)

        self._span_to_parent_span[span.id] = parent_span_id
        if parent_span_id is None:
            self._span_trees.append(span)

        self._span_to_trace[span.id] = trace_id
        self._span_observations[span.id] = span

    @property
    def span_trees(self) -> List[models.SpanModel]:
        self._build_spans_tree()
        return self._span_trees

    def _build_spans_tree(self) -> None:
        """
        Builds a list of span trees based on the data from the processed messages.
        Children's spans are sorted by creation time
        """
        with self._rlock:
            for span_id, parent_span_id in self._span_to_parent_span.items():
                if parent_span_id is None:
                    continue

                parent_span = self._span_observations[parent_span_id]
                if not _observation_already_stored(span_id, parent_span.spans):
                    parent_span.spans.append(self._span_observations[span_id])
                    parent_span.spans.sort(key=lambda x: x.start_time)

            all_span_ids = self._span_to_trace
            for span_id in all_span_ids:
                span = self._span_observations[span_id]
                span.feedback_scores = self._span_to_feedback_scores[span_id]

            self._span_trees.sort(key=lambda x: x.start_time)

    def _dispatch_message(self, message: messages.BaseMessage) -> None:
        message_type = type(message)
        handler = self._handlers.get(message_type)
        if handler is None:
            LOGGER.debug("Unknown type of message - %s", message_type.__name__)
            return

        handler(message)

    def process(self, message: messages.BaseMessage) -> None:
        with self._rlock:
            if not self.is_active():
                return

            try:
                self._dispatch_message(message)
            except Exception as exception:
                LOGGER.error(
                    "Failed to process message by emulator message processor, reason: %s",
                    exception,
                    exc_info=True,
                )

    @abc.abstractmethod
    def create_trace_model(
        self,
        trace_id: str,
        start_time: datetime.datetime,
        name: Optional[str],
        project_name: str,
        input: Any,
        output: Any,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        end_time: Optional[datetime.datetime],
        spans: Optional[List[models.SpanModel]],
        feedback_scores: Optional[List[models.FeedbackScoreModel]],
        error_info: Optional[ErrorInfoDict],
        thread_id: Optional[str],
        last_updated_at: Optional[datetime.datetime] = None,
    ) -> models.TraceModel:
        """
        Creates a trace model with the specified attributes. The method is abstract and must be
        implemented in a subclass to define how a trace model is created. It involves parameters
        such as timing details, trace-specific metadata, associated tags, input/output data,
        and other relevant information.

        Args:
            trace_id: A unique identifier for the trace.
            start_time: The starting time of the trace.
            name: An optional name representing the trace's purpose or identifier.
            project_name: The name of the project associated with the trace.
            input: Input data associated with the trace.
            output: Output data generated from the trace execution.
            tags: Optional list of tags for categorizing or labeling the trace.
            metadata: Optional dictionary containing additional metadata related
                to the trace.
            end_time: An optional datetime indicating when the trace ended.
            spans: A list of SpanModel instances representing spans of the trace.
            feedback_scores: Optional list of FeedbackScoreModel instances detailing
                collected feedback or evaluation metrics.
            error_info: Optional dictionary providing details on any errors
                encountered during the trace.
            thread_id: An optional identifier for the thread executing the trace.
            last_updated_at: Optional datetime to specify when the trace was
                last updated.

        Returns:
            An instance of TraceModel with the specified attributes.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
        """
        raise NotImplementedError("This method must be implemented in a subclass.")

    @abc.abstractmethod
    def create_span_model(
        self,
        span_id: str,
        start_time: datetime.datetime,
        name: Optional[str],
        input: Any,
        output: Any,
        tags: Optional[List[str]],
        metadata: Optional[Dict[str, Any]],
        type: SpanType,
        usage: Optional[Dict[str, Any]],
        end_time: Optional[datetime.datetime],
        project_name: str,
        spans: Optional[List[models.SpanModel]],
        feedback_scores: Optional[List[models.FeedbackScoreModel]],
        model: Optional[str],
        provider: Optional[str],
        error_info: Optional[ErrorInfoDict],
        total_cost: Optional[float],
        last_updated_at: Optional[datetime.datetime],
    ) -> models.SpanModel:
        """
        Abstract method to create a span model representing a span of a trace.
        This method is intended to facilitate the creation and organization of
        span-related data for tracing or debugging purposes.

        Args:
            span_id: Unique identifier for the span.
            start_time : The start timestamp of the span.
            name: Name of the span (e.g., an operation or process name).
            input: Input data or parameters related to the span.
            output: Output data or result associated with the span.
            tags: List of tags or labels to categorize the span.
            metadata: Additional metadata about the span.
            type: The type or category of the span.
            usage: Information about resource usage within the span.
            end_time: The end timestamp of the span.
            project_name: Name of the project to which the span belongs.
            spans: List of child or nested spans within the current span context.
            feedback_scores: Feedback scores related to this span, if applicable.
            model: The model identifier relevant to the span's processing.
            provider: The provider associated with the span or model.
            error_info: Information regarding errors encountered during the span's execution.
            total_cost: Total cost incurred during the span.
            last_updated_at: Timestamp marking the last update of the span's information.

        Returns:
            models.SpanModel: A fully initialized span model.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
        """
        raise NotImplementedError("This method must be implemented in a subclass.")

    @abc.abstractmethod
    def create_feedback_score_model(
        self,
        score_id: str,
        name: str,
        value: float,
        category_name: Optional[str],
        reason: Optional[str],
    ) -> models.FeedbackScoreModel:
        """
        Creates a feedback score model with the specified parameters.

        The method is abstract and must be implemented by a subclass. It defines
        the structure for creating a feedback score model by combining the given
        data into the appropriate model object.

        Args:
            score_id: The unique identifier for the feedback score.
            name: The name associated with the feedback score.
            value: The numeric value representing the feedback score.
            category_name: An optional category name for classifying the score.
            reason: An optional explanation or reason related to the feedback
                score.

        Returns:
            An instance of FeedbackScoreModel containing the created feedback
            score details.

        Raises:
            NotImplementedError: This method must be implemented in a subclass.
        """
        raise NotImplementedError("This method must be implemented in a subclass.")

    def _register_handlers(self) -> None:
        self._handlers: Dict[Type, Callable[[messages.BaseMessage], None]] = {
            messages.CreateSpanMessage: self._handle_create_span_message,  # type: ignore
            messages.CreateTraceMessage: self._handle_create_trace_message,  # type: ignore
            messages.UpdateSpanMessage: self._handle_update_span_message,  # type: ignore
            messages.UpdateTraceMessage: self._handle_update_trace_message,  # type: ignore
            messages.AddTraceFeedbackScoresBatchMessage: self._handle_add_trace_feedback_scores_batch_message,  # type: ignore
            messages.AddSpanFeedbackScoresBatchMessage: self._handle_add_span_feedback_scores_batch_message,  # type: ignore
            messages.CreateSpansBatchMessage: self._handle_create_spans_batch_message,  # type: ignore
            messages.CreateTraceBatchMessage: self._handle_create_traces_batch_message,  # type: ignore
            messages.CreateExperimentItemsBatchMessage: self._handle_create_experiment_items_batch_message,  # type: ignore
            messages.AttachmentSupportingMessage: self._noop_handler,  # type: ignore
        }

    def _handle_create_trace_message(
        self, message: messages.CreateTraceMessage
    ) -> None:
        trace = self.create_trace_model(
            trace_id=message.trace_id,
            name=message.name,
            input=message.input,
            output=message.output,
            tags=message.tags,
            metadata=message.metadata,
            start_time=message.start_time,
            end_time=message.end_time,
            project_name=message.project_name,
            error_info=message.error_info,
            thread_id=message.thread_id,
            last_updated_at=message.last_updated_at,
            feedback_scores=None,
            spans=None,
        )

        self._save_trace(trace)

    def _handle_create_span_message(self, message: messages.CreateSpanMessage) -> None:
        span = self.create_span_model(
            span_id=message.span_id,
            name=message.name,
            input=message.input,
            output=message.output,
            tags=message.tags,
            metadata=message.metadata,
            type=message.type,
            start_time=message.start_time,
            end_time=message.end_time,
            usage=message.usage,
            project_name=message.project_name,
            model=message.model,
            provider=message.provider,
            error_info=message.error_info,
            total_cost=message.total_cost,
            last_updated_at=message.last_updated_at,
            spans=None,
            feedback_scores=None,
        )

        self._save_span(
            span, trace_id=message.trace_id, parent_span_id=message.parent_span_id
        )

    def _handle_update_span_message(self, message: messages.UpdateSpanMessage) -> None:
        span = self._span_observations[message.span_id]
        update_payload = {
            "output": message.output,
            "usage": message.usage,
            "provider": message.provider,
            "model": message.model,
            "end_time": message.end_time,
            "metadata": message.metadata,
            "error_info": message.error_info,
            "tags": message.tags,
            "input": message.input,
            "total_cost": message.total_cost,
        }
        cleaned_update_payload = dict_utils.remove_none_from_dict(update_payload)
        span.__dict__.update(cleaned_update_payload)

    def _handle_update_trace_message(
        self, message: messages.UpdateTraceMessage
    ) -> None:
        current_trace = self._trace_observations[message.trace_id]
        update_payload = {
            "output": message.output,
            "end_time": message.end_time,
            "metadata": message.metadata,
            "error_info": message.error_info,
            "tags": message.tags,
            "input": message.input,
            "thread_id": message.thread_id,
        }
        cleaned_update_payload = dict_utils.remove_none_from_dict(update_payload)
        current_trace.__dict__.update(cleaned_update_payload)

    def _handle_add_span_feedback_scores_batch_message(
        self, message: messages.AddSpanFeedbackScoresBatchMessage
    ) -> None:
        for feedback_score_message in message.batch:
            feedback_model = self.create_feedback_score_model(
                score_id=feedback_score_message.id,
                name=feedback_score_message.name,
                value=feedback_score_message.value,
                category_name=feedback_score_message.category_name,
                reason=feedback_score_message.reason,
            )
            self._span_to_feedback_scores[feedback_score_message.id].append(
                feedback_model
            )

    def _handle_add_trace_feedback_scores_batch_message(
        self, message: messages.AddTraceFeedbackScoresBatchMessage
    ) -> None:
        for feedback_score_message in message.batch:
            feedback_model = self.create_feedback_score_model(
                score_id=feedback_score_message.id,
                name=feedback_score_message.name,
                value=feedback_score_message.value,
                category_name=feedback_score_message.category_name,
                reason=feedback_score_message.reason,
            )
            self._trace_to_feedback_scores[feedback_score_message.id].append(
                feedback_model
            )

    def _handle_create_spans_batch_message(
        self, message: messages.CreateSpansBatchMessage
    ) -> None:
        for item in message.batch:
            self._handle_span_write(item)

    def _handle_span_write(self, message: span_write.SpanWrite) -> None:
        if message.error_info is not None:
            error_info = ErrorInfoDict(
                exception_type=message.error_info.exception_type,
                message=message.error_info.message,
                traceback=message.error_info.traceback,
            )
        else:
            error_info = None

        span = self.create_span_model(
            span_id=message.id,
            name=message.name,
            input=message.input,
            output=message.output,
            tags=message.tags,
            metadata=message.metadata,
            type=message.type,
            start_time=message.start_time,
            end_time=message.end_time,
            usage=message.usage,
            project_name=message.project_name,
            model=message.model,
            provider=message.provider,
            total_cost=message.total_estimated_cost,
            last_updated_at=message.last_updated_at,
            spans=None,
            feedback_scores=None,
            error_info=error_info,
        )

        self._save_span(
            span, trace_id=message.trace_id, parent_span_id=message.parent_span_id
        )

    def _handle_create_traces_batch_message(
        self, message: messages.CreateTraceBatchMessage
    ) -> None:
        for item in message.batch:
            self._handle_trace_write(item)

    def _handle_trace_write(self, message: trace_write.TraceWrite) -> None:
        if message.error_info is not None:
            error_info = ErrorInfoDict(
                exception_type=message.error_info.exception_type,
                message=message.error_info.message,
                traceback=message.error_info.traceback,
            )
        else:
            error_info = None

        trace = self.create_trace_model(
            trace_id=message.id,
            name=message.name,
            input=message.input,
            output=message.output,
            tags=message.tags,
            metadata=message.metadata,
            start_time=message.start_time,
            end_time=message.end_time,
            project_name=message.project_name,
            thread_id=message.thread_id,
            last_updated_at=message.last_updated_at,
            spans=None,
            feedback_scores=None,
            error_info=error_info,
        )
        self._save_trace(trace)

    def _handle_create_experiment_items_batch_message(
        self, message: messages.CreateExperimentItemsBatchMessage
    ) -> None:
        for experiment_item_message in message.batch:
            experiment_item = models.ExperimentItemModel(
                id=experiment_item_message.id,
                experiment_id=experiment_item_message.experiment_id,
                trace_id=experiment_item_message.trace_id,
                dataset_item_id=experiment_item_message.dataset_item_id,
            )
            self._experiment_items.append(experiment_item)

    def _noop_handler(self, message: messages.BaseMessage) -> None:
        # just ignore the message
        pass

    @property
    def experiment_items(self) -> List[models.ExperimentItemModel]:
        """Returns the list of experiment items collected."""
        with self._rlock:
            return self._experiment_items


def _observation_already_stored(
    observation_id: str,
    observations: Union[List[models.SpanModel], List[models.TraceModel]],
) -> bool:
    for observation in observations:
        if observation.id == observation_id:
            return True

    return False



from typing import Any, Dict, TypedDict

from botocore import eventstream
import botocore.response


class ConverseStreamOutput(TypedDict):
    stream: eventstream.EventStream
    ResponseMetadata: Dict[str, Any]


class InvokeModelOutput(TypedDict):
    body: botocore.response.StreamingBody
    ResponseMetadata: Dict[str, Any]


class InvokeModelWithResponseStreamOutput(TypedDict):
    body: eventstream.EventStream
    ResponseMetadata: Dict[str, Any]



from typing import Dict, Any, Set, List, Literal

import opik.hooks

from .. import jsonable_encoder
from ..anonymizer import anonymizer


def encode_and_anonymize(
    kwargs_dict: Dict[str, Any],
    fields_to_anonymize: Set[str],
    object_type: Literal["span", "trace"],
) -> Dict[str, Any]:
    """
    Encodes and anonymizes the data in the given dictionary based on the specified
    fields using registered anonymizers. If no anonymizers are registered, the
    function simply encodes the dictionary without anonymization.

    Args:
        kwargs_dict: The dictionary containing the data to encode
            and anonymize.
        fields_to_anonymize: The set of fields within the dictionary to
            anonymize.
        object_type: A string indicating the type of object ('span' or 'trace')
            that was used to create the kwargs_dict. This is passed to anonymizers
            to provide context about the source object.

    Returns:
        A dictionary that has been encoded and, if applicable, anonymized.
    """
    # check if any anonymizer was registered
    encoded_obj = jsonable_encoder.encode(kwargs_dict)
    if not opik.hooks.has_anonymizers():
        return encoded_obj

    anonymizers = opik.hooks.get_anonymizers()
    return anonymize_encoded_obj(
        obj=encoded_obj,
        fields_to_anonymize=fields_to_anonymize,
        anonymizers=anonymizers,
        object_type=object_type,
    )


def anonymize_encoded_obj(
    obj: Dict[str, Any],
    fields_to_anonymize: Set[str],
    anonymizers: List[anonymizer.Anonymizer],
    object_type: Literal["span", "trace"],
) -> Dict[str, Any]:
    """
    Anonymizes specified fields in an encoded dictionary using the provided anonymizers.
    This function iterates over the given set of field names and applies each anonymizer
    to the corresponding field in the dictionary, if present. The anonymizers are expected
    to implement an `anonymize` method that takes the field value, field name, and object type
    as arguments. Only fields present in the dictionary and listed in `fields_to_anonymize`
    are anonymized.

    Args:
        obj: The encoded dictionary whose fields are to be anonymized.
        fields_to_anonymize: A set of field names within the dictionary to anonymize.
        anonymizers: A list of anonymizer instances to apply to each field.
        object_type: A string indicating the type of object ('span' or 'trace'),
            providing context for anonymization.

    Returns:
        The dictionary with specified fields anonymized using the provided anonymizers.
    """
    if isinstance(obj, dict):
        for field_name in fields_to_anonymize:
            if field_name in obj:
                for anonymizer_instance in anonymizers:
                    obj[field_name] = anonymizer_instance.anonymize(
                        obj[field_name],
                        field_name=field_name,
                        object_type=object_type,
                    )

    return obj



import numbers
from typing import Any, List, Optional, Tuple


class Parameter:
    __slots__ = ["name", "value", "types", "allow_empty", "possible_values"]

    def __init__(
        self,
        name: str,
        value: Any,
        types: Tuple,
        possible_values: Optional[List] = None,
        allow_empty: bool = True,
    ):
        self.name = name
        self.value = value
        self.types = types
        self.allow_empty = allow_empty
        self.possible_values = possible_values


def create_str_parameter(
    name: str,
    value: Any,
    possible_values: Optional[List[str]] = None,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(str,),
        possible_values=possible_values,
        allow_empty=allow_empty,
    )


def create_int_parameter(
    name: str,
    value: Any,
    possible_values: Optional[List[int]] = None,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(int,),
        possible_values=possible_values,
        allow_empty=allow_empty,
    )


def create_float_parameter(
    name: str,
    value: Any,
    possible_values: Optional[List[float]] = None,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(float,),
        possible_values=possible_values,
        allow_empty=allow_empty,
    )


def create_numeric_parameter(
    name: str,
    value: Any,
    possible_values: Optional[List[numbers.Number]] = None,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(float, int),
        possible_values=possible_values,
        allow_empty=allow_empty,
    )


def create_bool_parameter(
    name: str,
    value: Any,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(bool,),
        possible_values=None,
        allow_empty=allow_empty,
    )


def create_list_parameter(
    name: str,
    value: Any,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(list,),
        possible_values=None,
        allow_empty=allow_empty,
    )


def create_dict_parameter(
    name: str,
    value: Any,
    allow_empty: bool = True,
) -> Parameter:
    return Parameter(
        name=name,
        value=value,
        types=(dict,),
        possible_values=None,
        allow_empty=allow_empty,
    )



import logging
import functools

from typing import Callable, Any


def warning_decorator(message: str, logger: logging.Logger) -> Callable:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            logger.warning(message)
            return func(*args, **kwargs)

        return wrapper

    return decorator



import logging
from typing import Dict, Any, Tuple

from ... import _logging


LOGGER = logging.getLogger(__name__)
LANGGRAPH_OUTPUT_SIZE_THRESHOLD = 5000


def _extract_command_update(outputs: Dict[str, Any]) -> Dict[str, Any]:
    """Extract state updates from LangGraph Command objects.

    When a LangGraph node returns a Command, LangChain wraps it in {"output": Command(...)}.
    This function detects Command objects and extracts the update dict to properly log state changes.

    Args:
        outputs: The outputs dict from a LangChain Run.

    Returns:
        The extracted update dict if a Command is found, otherwise the original outputs.
    """
    if "output" in outputs and len(outputs) == 1:
        output_value = outputs["output"]
        # Duck-type check for Command object
        if hasattr(output_value, "update") and hasattr(output_value, "goto"):
            try:
                update_dict = output_value.update
                if isinstance(update_dict, dict):
                    _logging.log_once_at_level(
                        logging.DEBUG,
                        "Extracted state update from LangGraph Command object",
                        LOGGER,
                    )
                    return update_dict
            except Exception as e:
                LOGGER.warning(
                    f"Failed to extract update from Command-like object: {e}",
                    exc_info=True,
                )

    return outputs


def split_big_langgraph_outputs(
    outputs: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Split large LangGraph outputs to extract messages for thread display.

    Returns:
        tuple: (filtered_output_for_display, additional_metadata_for_span)

    LangGraph agents often produce complex outputs with large internal state
    that breaks thread display. This extracts conversational messages for
    clean thread display while preserving the full state in metadata.
    """
    if not isinstance(outputs, dict):
        return outputs, {}

    outputs = _extract_command_update(outputs)

    langgraph_like_output = "messages" in outputs and len(outputs) > 1
    if langgraph_like_output:
        output_str = str(outputs)
        output_size = len(output_str)

        if output_size > LANGGRAPH_OUTPUT_SIZE_THRESHOLD:
            _logging.log_once_at_level(
                logging.WARNING,
                f"Filtering large LangGraph output ({output_size} chars) for thread display",
                LOGGER,
            )

            filtered_output = {
                "messages": outputs["messages"],
            }

            if "thread_id" in outputs:
                filtered_output["thread_id"] = outputs["thread_id"]

            additional_metadata = {
                "_opik_langgraph_full_output": outputs,
                "_opik_output_filtering": {
                    "filtered": True,
                    "original_size_chars": output_size,
                    "filtered_keys": [
                        k for k in outputs.keys() if k not in ["messages", "thread_id"]
                    ],
                    "reason": "Large LangGraph output filtered for better thread display",
                },
            }

            return filtered_output, additional_metadata

    return outputs, {}



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .trace_filter_write import TraceFilterWrite
from .user_defined_metric_python_code_write import UserDefinedMetricPythonCodeWrite


class AutomationRuleEvaluatorUserDefinedMetricPythonWrite(UniversalBaseModel):
    filters: typing.Optional[typing.List[TraceFilterWrite]] = None
    code: typing.Optional[UserDefinedMetricPythonCodeWrite] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from .experiment import Experiment
from .helpers import build_metadata_and_prompt_versions, handle_prompt_args

__all__ = [
    "Experiment",
    "build_metadata_and_prompt_versions",
    "handle_prompt_args",
]



import copy
import logging
from typing import Any, Dict, List, Mapping, Optional, Tuple

from opik import id_helpers
import opik.jsonable_encoder as jsonable_encoder

from ..prompt import base_prompt


LOGGER = logging.getLogger(__name__)

PromptVersion = Dict[str, str]


def build_metadata_and_prompt_versions(
    experiment_config: Optional[Dict[str, Any]],
    prompts: Optional[List[base_prompt.BasePrompt]],
) -> Tuple[Optional[Dict[str, Any]], Optional[List[PromptVersion]]]:
    prompt_versions: Optional[List[PromptVersion]] = None

    if experiment_config is None:
        experiment_config = {}
    else:
        experiment_config = copy.deepcopy(experiment_config)

    if not isinstance(experiment_config, Mapping):
        LOGGER.error(
            "Experiment config must be dictionary, but %s was provided. Provided config will be ignored.",
            experiment_config,
        )
        experiment_config = {}

    if prompts is not None and len(prompts) > 0 and "prompts" in experiment_config:
        LOGGER.warning(
            "The `prompts` parameter will not be added to experiment since there is already `prompts` specified in experiment_config"
        )
        return experiment_config, None

    if prompts is not None and len(prompts) > 0:
        prompt_versions = []
        experiment_config["prompts"] = {}

        for prompt_obj in prompts:
            prompt_versions.append({"id": prompt_obj.__internal_api__version_id__})
            # Use __internal_api__to_info_dict__() to get the prompt content in a consistent way
            prompt_info = prompt_obj.__internal_api__to_info_dict__()
            # Extract the template/messages from the version dict
            if "version" in prompt_info:
                if "template" in prompt_info["version"]:
                    experiment_config["prompts"][prompt_obj.name] = prompt_info[
                        "version"
                    ]["template"]
                elif "messages" in prompt_info["version"]:
                    experiment_config["prompts"][prompt_obj.name] = prompt_info[
                        "version"
                    ]["messages"]

    if experiment_config == {}:
        return None, None

    metadata = jsonable_encoder.encode(experiment_config)

    return metadata, prompt_versions


def handle_prompt_args(
    prompt: Optional[base_prompt.BasePrompt] = None,
    prompts: Optional[List[base_prompt.BasePrompt]] = None,
) -> Optional[List[base_prompt.BasePrompt]]:
    if prompts is not None and len(prompts) > 0 and prompt is not None:
        LOGGER.warning(
            "Arguments `prompt` and `prompts` are mutually exclusive, `prompts` will be used`."
        )
    elif prompt is not None:
        prompts = [prompt]
    elif prompts is not None and len(prompts) == 0:
        prompts = None

    return prompts


def generate_unique_experiment_name(experiment_name_prefix: Optional[str]) -> str:
    if experiment_name_prefix is None:
        return id_helpers.generate_random_alphanumeric_string(12)

    return (
        f"{experiment_name_prefix}-{id_helpers.generate_random_alphanumeric_string(6)}"
    )
